@@ -16,10 +16,10 @@ from calendar import isleap
 from decimal import Decimal
 from typing import Union
 
-from .helpers import MONTH_DAYS_LEAP, MONTH_DAYS, DAYS_IN_4Y, \
+from ..helpers import MONTH_DAYS_LEAP, MONTH_DAYS, DAYS_IN_4Y, \
     DAYS_IN_100Y, DAYS_IN_400Y, days_from_common_era, adjust_day, \
     normalized_seconds, months2days, round_number
-from .atomic_types import AtomicTypeABCMeta, AnyAtomicType
+from .atomic_types import AtomicTypeMeta, AnyAtomicType
 from .untyped import UntypedAtomic
 
 
@@ -114,7 +114,7 @@ class Timezone(datetime.tzinfo):
                             "datetime.datetime instance or None")
 
 
-class AbstractDateTime(metaclass=AtomicTypeABCMeta):
+class AbstractDateTime(metaclass=AtomicTypeMeta):
     """
     A class for representing XSD date/time objects. It uses and internal datetime.datetime
     attribute and an integer attribute for processing BCE years or for years after 9999 CE.
@@ -253,21 +253,9 @@ class AbstractDateTime(metaclass=AtomicTypeABCMeta):
 
         if 'microsecond' in kwargs:
             microseconds = match.groupdict()['microsecond']
-            pow10 = 6 - len(microseconds)
-            try:
-                zeros = min(k for k in range(len(microseconds)) if '0' != microseconds[k])
-            except ValueError:
-                zeros = 0
-
-            if pow10 == 0:
-                pass
-            elif pow10 > 0:
-                kwargs['microsecond'] = kwargs['microsecond'] * 10**pow10
-            elif (kwargs['microsecond'] // 1 ** zeros) > 999999:
-                msg = "Invalid value {} for microsecond"
-                raise OverflowError(msg.format(kwargs['microsecond']))
-            else:
-                kwargs['microsecond'] = int(match.groupdict()['microsecond'][:6])
+            if len(microseconds) != 6:
+                microseconds += '0' * (6 - len(microseconds))
+                kwargs['microsecond'] = int(microseconds[:6])
 
         if 'year' in kwargs:
             year_digits = match.groupdict()['year'].lstrip('-')
@@ -721,7 +709,7 @@ class Duration(AnyAtomicType):
             raise ValueError('signs differ: (months=%d, seconds=%d)' % (months, seconds))
         elif abs(months) > 2 ** 31:
             raise OverflowError("months duration overflow")
-        elif abs(seconds) > 2 ** 63:
+        elif abs(seconds) > 2 ** 63:  # type: ignore[operator]
             raise OverflowError("seconds duration overflow")
 
         self.months = months
