@@ -17,6 +17,8 @@ from decimal import Decimal, DecimalException
 from itertools import takewhile
 from abc import ABCMeta
 from collections.abc import MutableSequence
+from typing import Dict, Type, Union, Optional
+
 
 #
 # Simple top down parser based on Vaughan Pratt's algorithm (Top Down Operator Precedence).
@@ -36,7 +38,7 @@ from collections.abc import MutableSequence
 # for a fully implementation example of a real parser.
 #
 
-# Parser special symbols set, that includes the TDOP's special symbols plus two
+# Parser special symbols set, that includes the special symbols of TDOP plus two
 # additional special symbols for managing invalid literals and unknown symbols.
 SPECIAL_SYMBOLS = frozenset((
     '(string)', '(float)', '(decimal)', '(integer)',
@@ -50,11 +52,11 @@ class ParseError(SyntaxError):
     """An error when parsing source with TDOP parser."""
 
 
-def count_leading_spaces(s):
+def count_leading_spaces(s: str) -> int:
     return sum(1 for _ in takewhile(str.isspace, s))
 
 
-def symbol_to_identifier(symbol):
+def symbol_to_classname(symbol: str) -> str:
     """
     Converts a symbol string to an identifier (only alphanumeric and '_').
     """
@@ -65,22 +67,22 @@ def symbol_to_identifier(symbol):
             return '%s_' % unicode_name(str(c)).title()
 
     if symbol.isalnum():
-        return symbol
+        return symbol.title()
     elif symbol in SPECIAL_SYMBOLS:
-        return symbol[1:-1]
+        return symbol[1:-1].title()
     elif all(c in '-_' for c in symbol):
-        value = '_'.join(unicode_name(str(c)).title() for c in symbol)
-        return value.replace(' ', '').replace('-', '')
+        value = ' '.join(unicode_name(c) for c in symbol)
+        return value.title().replace(' ', '').replace('-', '').replace('_', '')
 
     value = symbol.replace('-', '_')
     if value.isidentifier():
-        return value
-    else:
-        value = ''.join(get_id_name(c) for c in symbol).replace(' ', '').replace('-', '')
-        return value[:-1] if value.endswith('_') else value
+        return value.title().replace('_', '')
+
+    value = ''.join(get_id_name(c) for c in symbol)
+    return value.replace(' ', '').replace('-', '').replace('_', '')
 
 
-class MultiLabel(object):
+class MultiLabel:
     """
     Helper class for defining multi-value label for tokens. Useful when a symbol has more roles.
     A label of this type has equivalence with each of its values.
@@ -91,7 +93,7 @@ class MultiLabel(object):
         label == 'function'  # True
         label == 'operator'  # True
     """
-    def __init__(self, *values):
+    def __init__(self, *values: str):
         self.values = values
 
     def __eq__(self, other):
@@ -109,14 +111,14 @@ class MultiLabel(object):
     def __hash__(self):
         return hash(self.values)
 
-    def __contains__(self, item):
+    def __contains__(self, item: str):
         return any(item in v for v in self.values)
 
-    def startswith(self, string):
-        return any(v.startswith(string) for v in self.values)
+    def startswith(self, s: str):
+        return any(v.startswith(s) for v in self.values)
 
-    def endswith(self, string):
-        return any(v.endswith(string) for v in self.values)
+    def endswith(self, s: str):
+        return any(v.endswith(s) for v in self.values)
 
 
 class Token(MutableSequence):
@@ -153,17 +155,19 @@ class Token(MutableSequence):
     covers multiple roles (eg. as XPath function or axis). In those cases the definitive \
     role is defined at parse time (nud and/or led methods) after the token instance creation.
     """
-    symbol = None     # the token identifier, key in the token table.
-    lbp = 0           # left binding power
-    rbp = 0           # right binding power
-    pattern = None    # the token regex pattern, for building the tokenizer.
-    label = 'symbol'  # optional label
+    lbp = 0  # left binding power
+    rbp = 0  # right binding power
+    symbol: Optional[str] = None   # the token identifier
+    pattern: Optional[str] = None  # a custom regex pattern for building the tokenizer
+    label = 'symbol'               # optional label
+
+    __slots__ = '_items', 'parser', 'value', '_source', 'span'
 
     def __init__(self, parser, value=None):
         self._items = []
         self.parser = parser
         self.value = value if value is not None else self.symbol
-        self._source = parser.source
+        self._source: str = parser.source
         try:
             self.span = parser.match.span()
         except AttributeError:
@@ -212,17 +216,20 @@ class Token(MutableSequence):
     @property
     def tree(self):
         """Returns a tree representation string."""
-        symbol, length = self.symbol, len(self)
-        if symbol == '(name)':
-            return u'(%s)' % self.value
-        elif symbol in SPECIAL_SYMBOLS:
-            return u'(%r)' % self.value
-        elif symbol == '(':
-            return '()' if not self else self[0].tree
-        elif not length:
-            return u'(%s)' % symbol
+        if self.symbol == '(name)':
+            return '(%s)' % self.value
+        elif self.symbol in SPECIAL_SYMBOLS:
+            return '(%r)' % self.value
+        elif self.symbol == '(':
+            if not self:
+                return '()'
+            elif len(self) == 1:
+                return self[0].tree
+            return '(%s)' % ' '.join(item.tree for item in self)
+        elif not self:
+            return '(%s)' % self.symbol
         else:
-            return u'(%s %s)' % (symbol, ' '.join(item.tree for item in self))
+            return '(%s %s)' % (self.symbol, ' '.join(item.tree for item in self))
 
     @property
     def source(self):
@@ -239,26 +246,22 @@ class Token(MutableSequence):
             if not length:
                 return symbol
             elif length == 1:
-                return u'%s %s' % (symbol, self[0].source)
+                if 'postfix' in self.label:
+                    return '%s %s' % (self[0].source, symbol)
+                return '%s %s' % (symbol, self[0].source)
             elif length == 2:
-                return u'%s %s %s' % (self[0].source, symbol, self[1].source)
+                return '%s %s %s' % (self[0].source, symbol, self[1].source)
             else:
-                return u'%s %s' % (symbol, ' '.join(item.source for item in self))
+                return '%s %s' % (symbol, ' '.join(item.source for item in self))
 
     @property
     def position(self):
         """A tuple with the position of the token in terms of line and column."""
-        if not isinstance(self._source, (str, bytes)):
-            return None, None
-
         token_index = self.span[0]
         line = self._source[:token_index].count('\n') + 1
         if line == 1:
-            column = token_index + 1
-        else:
-            column = token_index - self._source[:token_index].rindex('\n') + 1
-
-        return line, column + count_leading_spaces(self._source[column - 1:])
+            return 1, token_index + 1
+        return line, token_index - self._source[:token_index].rindex('\n')
 
     def nud(self):
         """Pratt's null denotation method"""
@@ -328,15 +331,14 @@ class ParserMeta(type):
         # Avoids more parsers definitions for a single module
         for k, v in sys.modules[cls.__module__].__dict__.items():
             if isinstance(v, ParserMeta) and v.__module__ == cls.__module__:
-                raise RuntimeError("Multiple parser class definitions per "
-                                   "module is not permitted: %r" % cls)
+                raise RuntimeError("Multiple parser class definitions per module are not allowed")
 
         # Checks and initializes class attributes
         if not hasattr(cls, 'token_base_class'):
             cls.token_base_class = Token
         if not hasattr(cls, 'literals_pattern'):
             cls.literals_pattern = re.compile(
-                r"""'[^']*'|"[^"]*"|(?:\d+|\.\d+)(?:\.\d*)?(?:[Ee][+-]?\d+)?"""
+                    r"""'[^']*'|"[^"]*"|(?:\d+|\.\d+)(?:\.\d*)?(?:[Ee][+-]?\d+)?"""
             )
         if not hasattr(cls, 'name_pattern'):
             cls.name_pattern = re.compile(r'[A-Za-z0-9_]+')
@@ -373,7 +375,9 @@ class Parser(metaclass=ParserMeta):
     SYMBOLS = SPECIAL_SYMBOLS
     token_base_class = Token
     tokenizer = None
-    symbol_table = {}
+    symbol_table: Dict[str, Union[ABCMeta, Type[Token]]] = {}
+
+    __slots__ = 'source', 'tokens', 'match', 'token', 'next_token'
 
     def __init__(self):
         if self.tokenizer is None:
@@ -385,14 +389,9 @@ class Parser(metaclass=ParserMeta):
         self.next_token = None
 
     def __eq__(self, other):
-        if self.token_base_class != other.token_base_class:
-            return False
-        elif self.SYMBOLS != other.SYMBOLS:
-            return False
-        elif self.symbol_table != other.symbol_table:
-            return False
-        else:
-            return True
+        return self.token_base_class is other.token_base_class and \
+            self.SYMBOLS == other.SYMBOLS and \
+            self.symbol_table == other.symbol_table
 
     def parse(self, source):
         """
@@ -404,12 +403,12 @@ class Parser(metaclass=ParserMeta):
         """
         try:
             try:
-                self.source = source
                 self.tokens = iter(self.tokenizer.finditer(source))
             except TypeError as err:
                 token = self.symbol_table['(invalid)'](self, type(source))
                 raise token.wrong_syntax('invalid source type, {}'.format(err))
 
+            self.source = source
             self.advance()
             root_token = self.expression()
             self.next_token.expected('(end)')
@@ -444,12 +443,13 @@ class Parser(metaclass=ParserMeta):
             else:
                 literal, symbol, name, unknown = self.match.groups()
                 if symbol is not None:
-                    symbol = symbol.strip()
                     try:
                         self.next_token = self.symbol_table[symbol](self)
                     except KeyError:
-                        self.next_token = self.symbol_table['(unknown)'](self, symbol)
-                        raise self.next_token.wrong_syntax()
+                        if self.name_pattern.match(symbol) is None:
+                            self.next_token = self.symbol_table['(unknown)'](self, symbol)
+                            raise self.next_token.wrong_syntax()
+                        self.next_token = self.symbol_table['(name)'](self, symbol)
                     break
                 elif literal is not None:
                     if literal[0] in '\'"':
@@ -481,7 +481,7 @@ class Parser(metaclass=ParserMeta):
                     self.next_token = self.symbol_table['(unknown)'](self, unknown)
                     raise self.next_token.wrong_syntax()
                 elif str(self.match.group()).strip():
-                    msg = "Unexpected matching %r: not compatible tokenizer."
+                    msg = "unexpected matching %r: incompatible tokenizer"
                     raise RuntimeError(msg % self.match.group())
         return self.next_token
 
@@ -517,10 +517,10 @@ class Parser(metaclass=ParserMeta):
                     else:
                         try:
                             self.next_token = self.symbol_table[symbol](self)
+                            break
                         except KeyError:
                             self.next_token = self.symbol_table['(unknown)'](self)
                             raise self.next_token.wrong_syntax()
-                        break
                 else:
                     source_chunk.append(self.match.group())
         return ''.join(source_chunk)
@@ -567,8 +567,12 @@ class Parser(metaclass=ParserMeta):
         if self.token is None:
             return True
         token_index = self.token.span[0]
-        line_start = self.source[0:token_index].rindex('\n') + 1
-        return not bool(self.source[line_start:token_index].strip())
+        try:
+            line_start = self.source[:token_index].rindex('\n') + 1
+        except ValueError:
+            return not bool(self.source[:token_index].strip())
+        else:
+            return not bool(self.source[line_start:token_index].strip())
 
     def is_spaced(self, before=True, after=True):
         """
@@ -582,10 +586,11 @@ class Parser(metaclass=ParserMeta):
         """
         if self.token is None:
             return False
+
         start, end = self.token.span
-        if before and start > 0 and self.source[start - 1] in ' \t\n':
-            return True
         try:
+            if before and start > 0 and self.source[start - 1] in ' \t\n':
+                return True
             return after and self.source[end] in ' \t\n'
         except IndexError:
             return False
@@ -603,22 +608,10 @@ class Parser(metaclass=ParserMeta):
         :param kwargs: Optional attributes/methods for the token class.
         :return: A token class.
         """
-        def symbol_escape(s):
-            s = re.escape(s)
-            s.replace(r'\ ', r'\s+')
-
-            if s.isalpha():
-                s = r'\b%s\b(?![\-\.])' % s
-            elif s[-2:] == r'\(':
-                s = r'%s\s*%s' % (s[:-2], s[-2:])
-            elif s[-4:] == r'\:\:':
-                s = r'%s\s*%s' % (s[:-4], s[-4:])
-            return s
-
         try:
             try:
                 if ' ' in symbol:
-                    raise ValueError("%r: a symbol can't contains whitespaces." % symbol)
+                    raise ValueError("%r: a symbol can't contain whitespaces" % symbol)
             except TypeError:
                 assert isinstance(symbol, type) and issubclass(symbol, Token), \
                     "A %r subclass requested, not %r." % (Token, symbol)
@@ -635,18 +628,14 @@ class Parser(metaclass=ParserMeta):
                 raise NameError('%r is not a symbol of the parser %r.' % (symbol, cls))
 
             kwargs['symbol'] = symbol
-            if 'pattern' not in kwargs:
-                pattern = symbol_escape(symbol) if len(symbol) > 1 else re.escape(symbol)
-                kwargs['pattern'] = pattern
-
             label = kwargs.get('label', 'symbol')
             if isinstance(label, tuple):
                 label = kwargs['label'] = MultiLabel(*label)
 
-            token_class_name = "_{}_{}_token".format(
-                symbol_to_identifier(symbol), str(label).replace(' ', '_')
+            token_class_name = "_{}{}".format(
+                symbol_to_classname(symbol), str(label).title().replace(' ', '')
             )
-            token_class_bases = (getattr(cls, 'token_base_class', object),)
+            token_class_bases = kwargs.get('bases', (cls.token_base_class,))
             kwargs.update({
                 '__module__': cls.__module__,
                 '__qualname__': token_class_name,
@@ -669,12 +658,12 @@ class Parser(metaclass=ParserMeta):
         return token_class
 
     @classmethod
-    def unregister(cls, symbol):
+    def unregister(cls, symbol: str):
         """Unregister a token class from the symbol table."""
         del cls.symbol_table[symbol.strip()]
 
     @classmethod
-    def duplicate(cls, symbol, new_symbol, **kwargs):
+    def duplicate(cls, symbol: str, new_symbol: str, **kwargs):
         """Duplicate a token class with a new symbol."""
         token_class = cls.symbol_table[symbol]
         new_token_class = cls.register(new_symbol, **kwargs)
@@ -690,7 +679,7 @@ class Parser(metaclass=ParserMeta):
         def nud(self):
             return self
 
-        def evaluate(self, *args, **kwargs):
+        def evaluate(self, *_args, **_kwargs):
             return self.value
 
         return cls.register(symbol, label='literal', lbp=bp, evaluate=evaluate, nud=nud)
@@ -708,7 +697,7 @@ class Parser(metaclass=ParserMeta):
         def nud(self):
             self[:] = self.parser.expression(rbp=bp),
             return self
-        return cls.register(symbol, label='operator', lbp=bp, rbp=bp, nud=nud)
+        return cls.register(symbol, label='prefix operator', lbp=bp, rbp=bp, nud=nud)
 
     @classmethod
     def postfix(cls, symbol, bp=0):
@@ -716,7 +705,7 @@ class Parser(metaclass=ParserMeta):
         def led(self, left):
             self[:] = left,
             return self
-        return cls.register(symbol, label='operator', lbp=bp, rbp=bp, led=led)
+        return cls.register(symbol, label='postfix operator', lbp=bp, rbp=bp, led=led)
 
     @classmethod
     def infix(cls, symbol, bp=0):
@@ -743,9 +732,10 @@ class Parser(metaclass=ParserMeta):
         token_class = cls.register(symbol, label='operator', lbp=bp, rbp=bp)
 
         def bind(func):
-            assert callable(getattr(token_class, func.__name__, None)), \
-                "The name %r does not match with a callable of %r." % (func.__name__, token_class)
-            setattr(token_class, func.__name__, func)
+            method_name = func.__name__.partition('_')[0]
+            if not callable(getattr(token_class, method_name)):
+                raise TypeError(f"The attribute {method_name!r} is not a callable of {token_class}")
+            setattr(token_class, method_name, func)
             return func
         return bind
 
@@ -774,33 +764,38 @@ class Parser(metaclass=ParserMeta):
 
         :param symbol_table: a dictionary containing the token classes of the formal language.
         """
-        tokenizer_pattern_template = r"""
-            (%s) |       # Literals
-            (%s|[%s]) |  # Symbols
-            (%s) |       # Names
-            (\S) |       # Unknown symbols
-            \s+          # Skip extra spaces
-        """
-        patterns = [
-            t.pattern.replace('#', r'\#') for s, t in symbol_table.items()
-            if s not in SPECIAL_SYMBOLS
-        ]
-        string_patterns = []
         character_patterns = []
+        string_patterns = []
+        name_patterns = []
+        custom_patterns = set()
 
-        for p in patterns:
-            if ' ' in p:
-                raise ValueError('pattern %r contains spaces' % p)
-            length = len(p)
-            if length == 1 or length == 2 and p[0] == '\\':
-                character_patterns.append(p)
+        for symbol, token_class in symbol_table.items():
+            if symbol in SPECIAL_SYMBOLS:
+                continue
+            elif token_class.pattern is not None:
+                custom_patterns.add(token_class.pattern)
+            elif cls.name_pattern.match(symbol) is not None:
+                name_patterns.append(re.escape(symbol))
+            elif len(symbol) == 1:
+                character_patterns.append(re.escape(symbol))
             else:
-                string_patterns.append(p)
+                string_patterns.append(re.escape(symbol))
 
-        pattern = tokenizer_pattern_template % (
+        symbols_patterns = []
+        if string_patterns:
+            symbols_patterns.append('|'.join(sorted(string_patterns, key=lambda x: -len(x))))
+        if character_patterns:
+            symbols_patterns.append('[{}]'.format(''.join(character_patterns)))
+        if name_patterns:
+            symbols_patterns.append(r'\b(?:{})\b(?![\-\.])'.format(
+                '|'.join(sorted(name_patterns, key=lambda x: -len(x)))
+            ))
+        if custom_patterns:
+            symbols_patterns.append('|'.join(custom_patterns))
+
+        tokenizer_pattern = r"({})|({})|({})|(\S)|\s+".format(
             cls.literals_pattern.pattern,
-            '|'.join(sorted(string_patterns, key=lambda x: -len(x))),
-            ''.join(character_patterns),
+            '|'.join(symbols_patterns),
             cls.name_pattern.pattern
         )
-        return re.compile(pattern, re.VERBOSE)
+        return re.compile(tokenizer_pattern)
