@@ -12,18 +12,21 @@ import math
 import decimal
 import operator
 from copy import copy
+from typing import Dict, Tuple
 
-from ..exceptions import MissingContextError, ElementPathKeyError, xpath_error
+from ..helpers import EQNAME_PATTERN, normalize_sequence_type
+from ..exceptions import MissingContextError, ElementPathKeyError, \
+    ElementPathValueError, xpath_error
 from ..datatypes import AnyAtomicType, AbstractDateTime, Duration, DayTimeDuration, \
     YearMonthDuration, NumericProxy, ArithmeticProxy, UntypedAtomic, QName, \
     xsd10_atomic_types, xsd11_atomic_types, ATOMIC_VALUES
 from ..xpath_context import XPathSchemaContext
 from ..tdop import Parser
 from ..namespaces import XML_NAMESPACE, XSD_NAMESPACE, XPATH_FUNCTIONS_NAMESPACE, \
-    XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE, XSD_UNTYPED_ATOMIC, get_namespace, \
-    get_expanded_name, split_expanded_name
+    XPATH_MATH_FUNCTIONS_NAMESPACE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE, \
+    XSD_UNTYPED_ATOMIC, get_namespace, get_expanded_name, split_expanded_name
 from ..schema_proxy import AbstractSchemaProxy
-from ..xpath_token import XPathToken
+from ..xpath_token import XPathToken, XPathAxis, XPathFunction
 from ..xpath_nodes import XPathNode, TypedElement, AttributeNode, TypedAttribute, \
     is_xpath_node, match_element_node, is_schema_node, is_document_node, \
     match_attribute_node, is_element_node, node_kind
@@ -96,7 +99,7 @@ class XPath1Parser(Parser):
     """
 
     # Labels and symbols admitted after a path step
-    PATH_STEP_LABELS = ('axis', 'kind test')
+    PATH_STEP_LABELS: Tuple[str, ...] = ('axis', 'kind test')
     PATH_STEP_SYMBOLS = {
         '(integer)', '(string)', '(float)', '(decimal)', '(name)', '*', '@', '..', '.', '{'
     }
@@ -106,6 +109,14 @@ class XPath1Parser(Parser):
     variable_types = None   # XPath 1.0 doesn't have in-scope variable types
     xsd_version = '1.0'     # Use XSD 1.0 datatypes for default
     function_namespace = XPATH_FUNCTIONS_NAMESPACE
+    function_signatures: Dict[Tuple[QName, int], str] = {}
+
+    # https://www.w3.org/TR/xpath-3/#id-reserved-fn-names
+    RESERVED_FUNCTION_NAMES = {
+        'array', 'attribute', 'comment', 'document-node', 'element', 'empty-sequence',
+        'function', 'if', 'item', 'map', 'namespace-node', 'node', 'processing-instruction',
+        'schema-attribute', 'schema-element', 'switch', 'text', 'typeswitch',
+    }
 
     def __init__(self, namespaces=None, strict=True, *args, **kwargs):
         super(XPath1Parser, self).__init__()
@@ -148,89 +159,46 @@ class XPath1Parser(Parser):
     @classmethod
     def axis(cls, symbol, reverse_axis=False, bp=80):
         """Register a token for a symbol that represents an XPath *axis*."""
-        def nud_(self):
-            self.parser.advance('::')
-            self.parser.expected_name(
-                '(name)', '*', 'text', 'node', 'document-node',
-                'comment', 'processing-instruction', 'attribute',
-                'schema-attribute', 'element', 'schema-element'
-            )
-            self[:] = self.parser.expression(rbp=bp),
-            return self
-
-        pattern = r'\b%s(?=\s*\:\:|\s*\(\:.*\:\)\s*\:\:)' % symbol
-        return cls.register(
-            symbol, pattern=pattern, label='axis',
-            reverse_axis=reverse_axis, lbp=bp, rbp=bp, nud=nud_
-        )
+        return cls.register(symbol, label='axis', bases=(XPathAxis,),
+                            reverse_axis=reverse_axis, lbp=bp, rbp=bp)
 
     @classmethod
-    def function(cls, symbol, nargs=None, label='function', bp=90):
+    def function(cls, symbol, nargs=None, sequence_types=(), label='function', bp=90):
         """
-        Registers a token class for a symbol that represents an XPath *callable* object.
-        For default a callable labeled as *function* is registered but a different label
-        can be provided.
+        Registers a token class for a symbol that represents an XPath function.
         """
-        def nud_(self):
-            code = 'XPST0017' if self.label == 'function' else 'XPST0003'
-            self.value = None
-            self.parser.advance('(')
-            if nargs is None:
-                del self[:]
-                if self.parser.next_token.symbol in (')', '(end)'):
-                    raise self.error(code, 'at least an argument is required')
-                while True:
-                    self.append(self.parser.expression(5))
-                    if self.parser.next_token.symbol != ',':
-                        break
-                    self.parser.advance()
-                self.parser.advance(')')
-                return self
-            elif nargs == 0:
-                if self.parser.next_token.symbol != ')':
-                    if self.parser.next_token.symbol != '(end)':
-                        raise self.error(code, '%s has no arguments' % str(self))
-                    raise self.parser.next_token.wrong_syntax()
-                self.parser.advance()
-                return self
-            elif isinstance(nargs, (tuple, list)):
-                min_args, max_args = nargs
+        if 'function' not in label:
+            pass
+        elif symbol in cls.RESERVED_FUNCTION_NAMES:
+            raise ElementPathValueError(f'{symbol!r} is a reserved function name')
+        elif sequence_types:
+            # Register function signature(s)
+            if label == 'math function':
+                qname = QName(XPATH_MATH_FUNCTIONS_NAMESPACE, 'math:%s' % symbol)
             else:
-                min_args = max_args = nargs
+                qname = QName(XPATH_FUNCTIONS_NAMESPACE, 'fn:%s' % symbol)
 
-            k = 0
-            while k < min_args:
-                if self.parser.next_token.symbol in (')', '(end)'):
-                    msg = 'Too few arguments: expected at least %s arguments' % min_args
-                    raise self.wrong_nargs(msg if min_args > 1 else msg[:-1])
+            if nargs is None:
+                pass  # pragma: no cover
+            elif isinstance(nargs, int):
+                assert len(sequence_types) == nargs + 1
+                cls.function_signatures[(qname, nargs)] = 'function({}) as {}'.format(
+                    ', '.join(sequence_types[:-1]), sequence_types[-1]
+                )
+            elif nargs[1] is None:
+                assert len(sequence_types) == nargs[0] + 1
+                cls.function_signatures[(qname, nargs[0])] = 'function({}, ...) as {}'.format(
+                    ', '.join(sequence_types[:-1]), sequence_types[-1]
+                )
+            else:
+                assert len(sequence_types) == nargs[1] + 1
+                for arity in range(nargs[0], nargs[1] + 1):
+                    cls.function_signatures[(qname, arity)] = 'function({}) as {}'.format(
+                        ', '.join(sequence_types[:arity]), sequence_types[-1]
+                    )
 
-                self[k:] = self.parser.expression(5),
-                k += 1
-                if k < min_args:
-                    if self.parser.next_token.symbol == ')':
-                        msg = 'Too few arguments: expected at least %s arguments' % min_args
-                        raise self.error(code, msg if min_args > 1 else msg[:-1])
-                    self.parser.advance(',')
-
-            while max_args is None or k < max_args:
-                if self.parser.next_token.symbol == ',':
-                    self.parser.advance(',')
-                    self[k:] = self.parser.expression(5),
-                elif k == 0 and self.parser.next_token.symbol != ')':
-                    self[k:] = self.parser.expression(5),
-                else:
-                    break  # pragma: no cover
-                k += 1
-
-            if self.parser.next_token.symbol == ',':
-                msg = 'Too many arguments: expected at most %s arguments' % max_args
-                raise self.error(code, msg if max_args > 1 else msg[:-1])
-
-            self.parser.advance(')')
-            return self
-
-        pattern = r'\b%s(?=\s*\(|\s*\(\:.*\:\)\()' % symbol
-        return cls.register(symbol, pattern=pattern, label=label, lbp=bp, rbp=bp, nud=nud_)
+        return cls.register(symbol, nargs=nargs, sequence_types=sequence_types, label=label,
+                            bases=(XPathFunction,), lbp=bp, rbp=bp)
 
     def parse(self, source):
         root_token = super(XPath1Parser, self).parse(source)
@@ -289,20 +257,66 @@ class XPath1Parser(Parser):
     def is_sequence_type(self, value):
         """Checks is a string is a sequence type specification."""
         try:
-            value = value.strip()
-        except AttributeError:
+            value = normalize_sequence_type(value)
+        except TypeError:
             return False
 
         if not value:
             return False
-        elif value == 'empty-sequence()':
+        elif value == 'empty-sequence()' or value == 'none':
             return True
-        elif value[-1] in ('?', '+', '*'):
+        elif value[-1] in {'?', '+', '*'}:
             value = value[:-1]
 
-        if value in {'untypedAtomic', 'attribute()', 'element()', 'text()', 'document-node()',
-                     'comment()', 'processing-instruction()', 'item()', 'node()'}:
+        if value in {'untypedAtomic', 'attribute()', 'attribute(*)', 'element()',
+                     'element(*)', 'text()', 'document-node()', 'comment()',
+                     'processing-instruction()', 'item()', 'node()', 'numeric'}:
             return True
+
+        elif value.startswith('element(') and value.endswith(')'):
+            if ',' not in value:
+                return EQNAME_PATTERN.match(value[8:-1]) is not None
+
+            try:
+                arg1, arg2 = value[8:-1].split(', ')
+            except ValueError:
+                return False
+            else:
+                return (arg1 == '*' or EQNAME_PATTERN.match(arg1) is not None) \
+                    and EQNAME_PATTERN.match(arg2) is not None
+
+        elif value.startswith('document-node(') and value.endswith(')'):
+            if not value.startswith('document-node(element('):
+                return False
+            return self.is_sequence_type(value[14:-1])
+
+        elif value.startswith('function('):
+            if value == 'function(*)' and self.version >= '3.0':
+                return True
+
+            try:
+                value, return_type = value.rsplit(' as ', 1)
+            except ValueError:
+                return False
+            else:
+                if not self.is_sequence_type(return_type):
+                    return False
+                elif value == 'function()':
+                    return True
+
+                value = value[9:-1]
+                if value.endswith(', ...'):
+                    value = value[:-5]
+
+                if 'function(' not in value:
+                    return all(self.is_sequence_type(x) for x in value.split(', '))
+
+                # Cover only if function() spec is the last argument
+                k = value.index('function(')
+                if not self.is_sequence_type(value[k:]):
+                    return False
+                return all(self.is_sequence_type(x) for x in value[:k].split(', ') if x)
+
         elif QName.pattern.match(value) is None:
             return False
 
@@ -357,7 +371,7 @@ class XPath1Parser(Parser):
         """
         if sequence_type[-1] in {'?', '+', '*'}:
             return self.match_sequence_type(value, sequence_type[:-1], sequence_type[-1])
-        elif value is None or value == []:
+        elif value is None or isinstance(value, list) and value == []:
             return sequence_type == 'empty-sequence()' or occurrence in {'?', '*'}
         elif sequence_type == 'empty-sequence()':
             return False
@@ -370,6 +384,8 @@ class XPath1Parser(Parser):
                 return all(self.match_sequence_type(x, sequence_type) for x in value)
         elif sequence_type == 'item()':
             return is_xpath_node(value) or isinstance(value, (AnyAtomicType, list))
+        elif sequence_type == 'numeric':
+            return isinstance(value, NumericProxy)
 
         value_kind = node_kind(value)
         if value_kind is not None:
@@ -384,15 +400,10 @@ class XPath1Parser(Parser):
     def check_variables(self, values):
         """Checks the sequence types of the XPath dynamic context's variables."""
         for varname, value in values.items():
-            if isinstance(value, list):
-                if self.match_sequence_type(value, 'item()', occurrence='*'):
-                    continue
-            else:
-                if self.match_sequence_type(value, 'item()'):
-                    continue
-
-            message = "Unmatched sequence type for variable {!r}".format(varname)
-            raise xpath_error('XPDY0050', message)
+            if not self.match_sequence_type(
+                    value, 'item()', occurrence='*' if isinstance(value, list) else None):
+                message = "Unmatched sequence type for variable {!r}".format(varname)
+                raise xpath_error('XPDY0050', message)
 
 
 ##
@@ -428,23 +439,33 @@ literal('(unknown)')
 
 
 @method(register('(name)', bp=10, label='literal'))
-def nud(self):
+def nud_name_literal(self):
     if self.parser.next_token.symbol == '(':
-        if self.namespace == XSD_NAMESPACE:
+        if self.parser.version >= '3.0':
+            pass
+        elif self.namespace == XSD_NAMESPACE:
             raise self.error('XPST0017', 'unknown constructor function {!r}'.format(self.value))
-        raise self.error('XPST0017', 'unknown function {!r}'.format(self.value))
+        elif self.value not in self.parser.RESERVED_FUNCTION_NAMES:
+            raise self.error('XPST0017', 'unknown function {!r}'.format(self.value))
+        elif self.value == 'typeswitch':
+            msg = 'improper use of XQuery reserved name {!r}'
+            raise self.error('XPST0003', msg.format(self.value))
+        else:
+            msg = 'improper use of XPath reserved name {!r}'
+            raise self.error('XPST0017', msg.format(self.value))
+
     elif self.parser.next_token.symbol == '::':
         raise self.missing_axis("axis '%s::' not found" % self.value)
     return self
 
 
 @method('(name)')
-def evaluate(self, context=None):
+def evaluate_name_literal(self, context=None):
     return [x for x in self.select(context)]
 
 
 @method('(name)')
-def select(self, context=None):
+def select_name_literal(self, context=None):
     if context is None:
         raise self.missing_context()
 
@@ -501,7 +522,7 @@ def select(self, context=None):
 ###
 # Namespace prefix reference
 @method(':', bp=95)
-def led(self, left):
+def led_namespace_prefix(self, left):
     if self.parser.version == '1.0':
         left.expected('(name)')
     else:
@@ -516,6 +537,8 @@ def led(self, left):
         try:
             namespace = self.get_namespace(left.value)
         except ElementPathKeyError:
+            self.parser.advance()  # Assure there isn't a following incomplete comment
+            self[:] = left, self.parser.token
             msg = "prefix {!r} is not declared".format(left.value)
             raise self.error('XPST0081', msg) from None
         else:
@@ -525,18 +548,22 @@ def led(self, left):
 
     self[:] = left, self.parser.expression(90)
     self.value = '{}:{}'.format(self[0].value, self[1].value)
+
+    if self.parser.next_token.symbol == ':':
+        raise self.wrong_syntax()
+
     return self
 
 
 @method(':')
-def evaluate(self, context=None):
+def evaluate_namespace_prefix(self, context=None):
     if self[1].label.endswith('function'):
         return self[1].evaluate(context)
     return [x for x in self.select(context)]
 
 
 @method(':')
-def select(self, context=None):
+def select_namespace_prefix(self, context=None):
     if self[1].label.endswith('function'):
         value = self[1].evaluate(context)
         if isinstance(value, list):
@@ -590,8 +617,8 @@ def select(self, context=None):
 ###
 # Namespace URI as in ElementPath
 @method('{', bp=95)
-def nud(self):
-    if self.parser.strict:
+def nud_namespace_uri(self):
+    if self.parser.strict and self.symbol == '{':
         raise self.wrong_syntax("not allowed symbol if parser has strict=True")
 
     namespace = self.parser.next_token.value + self.parser.advance_until('}')
@@ -602,38 +629,42 @@ def nud(self):
 
     self[:] = self.parser.symbol_table['(string)'](self.parser, namespace), \
         self.parser.expression(90)
+
+    if self[1].value is None:
+        self.value = None
+    else:
+        self.value = '{%s}%s' % (self[0].value, self[1].value)
     return self
 
 
 @method('{')
-def evaluate(self, context=None):
+def evaluate_namespace_uri(self, context=None):
     if self[1].label.endswith('function'):
         return self[1].evaluate(context)
     return [x for x in self.select(context)]
 
 
 @method('{')
-def select(self, context=None):
+def select_namespace_uri(self, context=None):
     if self[1].label.endswith('function'):
         yield self[1].evaluate(context)
         return
     elif context is None:
         raise self.missing_context()
 
-    name = '{%s}%s' % (self[0].value, self[1].value)
     if isinstance(context, XPathSchemaContext):
-        yield from self.select_xsd_nodes(context, name)
+        yield from self.select_xsd_nodes(context, self.value)
 
     elif self.xsd_types is None:
         for item in context.iter_children_or_self():
-            if match_attribute_node(item, name):
-                yield item.value
-            elif match_element_node(item, name):
+            if match_attribute_node(item, self.value):
+                yield item
+            elif match_element_node(item, self.value):
                 yield item
     else:
         # XSD typed selection
         for item in context.iter_children_or_self():
-            if match_attribute_node(item, name) or match_element_node(item, name):
+            if match_attribute_node(item, self.value) or match_element_node(item, self.value):
                 if isinstance(item, (TypedAttribute, TypedElement)):
                     yield item
                 else:
@@ -644,7 +675,7 @@ def select(self, context=None):
 ###
 # Variables
 @method('$', bp=90)
-def nud(self):
+def nud_variable_reference(self):
     self.parser.expected_name('(name)')
     self[:] = self.parser.expression(rbp=90),
     if ':' in self[0].value:
@@ -653,7 +684,7 @@ def nud(self):
 
 
 @method('$')
-def evaluate(self, context=None):
+def evaluate_variable_reference(self, context=None):
     if context is None:
         raise self.missing_context()
 
@@ -666,7 +697,7 @@ def evaluate(self, context=None):
 ###
 # Nullary operators (use only the context)
 @method(nullary('*'))
-def select(self, context=None):
+def select_wildcard(self, context=None):
     if self:
         # Product operator
         item = self.evaluate(context)
@@ -690,7 +721,7 @@ def select(self, context=None):
                 pass  # '*' wildcard doesn't match document nodes
             elif context.axis == 'attribute':
                 if isinstance(item, (AttributeNode, TypedAttribute)):
-                    yield item.value
+                    yield item
             elif is_element_node(item):
                 yield item
 
@@ -706,7 +737,7 @@ def select(self, context=None):
 
 
 @method(nullary('.'))
-def select(self, context=None):
+def select_self_shortcut(self, context=None):
     if context is None:
         raise self.missing_context()
 
@@ -740,7 +771,7 @@ def select(self, context=None):
 
 
 @method(nullary('..'))
-def select(self, context=None):
+def select_parent_shortcut(self, context=None):
     if context is None:
         raise self.missing_context()
     else:
@@ -753,13 +784,13 @@ def select(self, context=None):
 ###
 # Logical Operators
 @method(infix('or', bp=20))
-def evaluate(self, context=None):
+def evaluate_or_operator(self, context=None):
     return self.boolean_value(self[0].evaluate(copy(context))) or \
         self.boolean_value(self[1].evaluate(copy(context)))
 
 
 @method(infix('and', bp=25))
-def evaluate(self, context=None):
+def evaluate_and_operator(self, context=None):
     return self.boolean_value(self[0].evaluate(copy(context))) and \
         self.boolean_value(self[1].evaluate(copy(context)))
 
@@ -772,7 +803,7 @@ def evaluate(self, context=None):
 @method('>', bp=30)
 @method('<=', bp=30)
 @method('>=', bp=30)
-def led(self, left):
+def led_comparison_operators(self, left):
     if left.symbol in OPERATORS_MAP:
         raise self.wrong_syntax()
     self[:] = left, self.parser.expression(rbp=30)
@@ -785,7 +816,7 @@ def led(self, left):
 @method('>')
 @method('<=')
 @method('>=')
-def evaluate(self, context=None):
+def evaluate_comparison_operators(self, context=None):
     op = OPERATORS_MAP[self.symbol]
     try:
         return any(op(x1, x2) for x1, x2 in self.iter_comparison_data(context))
@@ -798,7 +829,7 @@ def evaluate(self, context=None):
 ###
 # Numerical operators
 @method(infix('+', bp=40))
-def evaluate(self, context=None):
+def evaluate_plus_operator(self, context=None):
     if len(self) == 1:
         arg = self.get_argument(context, cls=NumericProxy)
         if arg is not None:
@@ -820,7 +851,7 @@ def evaluate(self, context=None):
 
 
 @method(infix('-', bp=40))
-def evaluate(self, context=None):
+def evaluate_minus_operator(self, context=None):
     if len(self) == 1:
         arg = self.get_argument(context, cls=NumericProxy)
         if arg is not None:
@@ -843,13 +874,13 @@ def evaluate(self, context=None):
 
 @method('+')
 @method('-')
-def nud(self):
+def nud_plus_minus_operators(self):
     self[:] = self.parser.expression(rbp=70),
     return self
 
 
 @method(infix('*', bp=45))
-def evaluate(self, context=None):
+def evaluate_multiply_operator(self, context=None):
     if self:
         op1, op2 = self.get_operands(context, cls=ArithmeticProxy)
         if op1 is not None:
@@ -886,7 +917,7 @@ def evaluate(self, context=None):
 
 
 @method(infix('div', bp=45))
-def evaluate(self, context=None):
+def evaluate_div_operator(self, context=None):
     dividend, divisor = self.get_operands(context, cls=ArithmeticProxy)
     if dividend is None:
         return
@@ -921,7 +952,7 @@ def evaluate(self, context=None):
 
 
 @method(infix('mod', bp=45))
-def evaluate(self, context=None):
+def evaluate_mod_operator(self, context=None):
     op1, op2 = self.get_operands(context, cls=NumericProxy)
     if op1 is not None:
         if op2 == 0 and isinstance(op2, float):
@@ -939,10 +970,20 @@ def evaluate(self, context=None):
             raise self.error('FOAR0001') from None
 
 
+# Resolve the intrinsic ambiguity of some infix operators
+@method('or')
+@method('and')
+@method('div')
+@method('mod')
+def nud_logical_div_mod_operators(self):
+    token = self.parser.symbol_table['(name)'](self.parser, self.symbol)
+    return token.nud()
+
+
 ###
 # Union expressions
 @method('|', bp=50)
-def led(self, left):
+def led_union_operator(self, left):
     self.cut_and_sort = True
     if left.symbol in {'|', 'union'}:
         left.cut_and_sort = False
@@ -951,11 +992,11 @@ def led(self, left):
 
 
 @method('|')
-def select(self, context=None):
+def select_union_operator(self, context=None):
     if context is None:
         raise self.missing_context()
 
-    results = {item for k in range(2) for item in self[k].select(context.copy())}
+    results = {item for k in range(2) for item in self[k].select(copy(context))}
     if any(not is_xpath_node(x) for x in results):
         raise self.error('XPTY0004', 'only XPath nodes are allowed')
     elif not self.cut_and_sort:
@@ -967,7 +1008,7 @@ def select(self, context=None):
 ###
 # Path expressions
 @method('//', bp=75)
-def nud(self):
+def nud_descendant_path(self):
     if self.parser.next_token.label not in self.parser.PATH_STEP_LABELS:
         self.parser.expected_name(*self.parser.PATH_STEP_SYMBOLS)
 
@@ -976,7 +1017,7 @@ def nud(self):
 
 
 @method('/', bp=75)
-def nud(self):
+def nud_child_path(self):
     if self.parser.next_token.label not in self.parser.PATH_STEP_LABELS:
         try:
             self.parser.expected_name(*self.parser.PATH_STEP_SYMBOLS)
@@ -989,7 +1030,7 @@ def nud(self):
 
 @method('//')
 @method('/')
-def led(self, left):
+def led_child_or_descendant_path(self, left):
     if self.parser.next_token.label not in self.parser.PATH_STEP_LABELS:
         self.parser.expected_name(*self.parser.PATH_STEP_SYMBOLS)
 
@@ -998,7 +1039,7 @@ def led(self, left):
 
 
 @method('/')
-def select(self, context=None):
+def select_child_path(self, context=None):
     """
     Child path expression. Selects child:: axis as default (when bind to '*' or '(name)').
     """
@@ -1039,7 +1080,7 @@ def select(self, context=None):
 
 
 @method('//')
-def select(self, context=None):
+def select_descendant_path(self, context=None):
     # Note: // is short for /descendant-or-self::node()/, so the axis
     #   is left to None. Use descendant:: only if next-step uses child
     #   axis, to preserve document order.
@@ -1066,14 +1107,14 @@ def select(self, context=None):
 ###
 # Predicate filters
 @method('[', bp=80)
-def led(self, left):
+def led_predicate(self, left):
     self[:] = left, self.parser.expression()
     self.parser.advance(']')
     return self
 
 
 @method('[')
-def select(self, context=None):
+def select_predicate(self, context=None):
     if context is None:
         raise self.missing_context()
 
@@ -1082,7 +1123,7 @@ def select(self, context=None):
                 and not is_xpath_node(context.item):
             raise self.error('XPTY0020')
 
-        predicate = [x for x in self[1].select(context.copy())]
+        predicate = [x for x in self[1].select(copy(context))]
         if len(predicate) == 1 and isinstance(predicate[0], NumericProxy):
             if context.position == predicate[0]:
                 yield context.item
@@ -1093,17 +1134,17 @@ def select(self, context=None):
 ###
 # Parenthesized expressions
 @method('(', bp=100)
-def nud(self):
+def nud_parenthesized_expr(self):
     self[:] = self.parser.expression(),
     self.parser.advance(')')
     return self
 
 
 @method('(')
-def evaluate(self, context=None):
+def evaluate_parenthesized_expr(self, context=None):
     return self[0].evaluate(context)
 
 
 @method('(')
-def select(self, context=None):
+def select_parenthesized_expr(self, context=None):
     return self[0].select(context)
