@@ -19,15 +19,16 @@ from copy import copy
 from decimal import Decimal, DivisionByZero
 from urllib.parse import urlparse
 
+from ..helpers import normalize_sequence_type
 from ..exceptions import ElementPathError, ElementPathTypeError, \
     ElementPathValueError, MissingContextError, xpath_error
 from ..namespaces import XSD_NAMESPACE, XML_NAMESPACE, XLINK_NAMESPACE, \
     XPATH_FUNCTIONS_NAMESPACE, XQT_ERRORS_NAMESPACE, XSD_NOTATION, \
     XSD_ANY_ATOMIC_TYPE, get_namespace, get_prefixed_name, get_expanded_name
-from ..datatypes import UntypedAtomic, QName, AnyURI, Duration
-from ..xpath_nodes import TypedAttribute, is_xpath_node, \
+from ..datatypes import UntypedAtomic, QName, AnyURI, Duration, Integer
+from ..xpath_nodes import TypedElement, is_xpath_node, \
     match_attribute_node, is_element_node, is_document_node
-from ..xpath_token import UNICODE_CODEPOINT_COLLATION
+from ..xpath_token import UNICODE_CODEPOINT_COLLATION, XPathFunction
 from ..xpath1 import XPath1Parser
 from ..xpath_context import XPathSchemaContext
 from ..schema_proxy import AbstractSchemaProxy
@@ -187,6 +188,8 @@ class XPath2Parser(XPath1Parser):
         '(integer)', '(string)', '(float)', '(decimal)', '(name)', '*', '@', '..', '.', '(', '{'
     }
 
+    function_signatures = XPath1Parser.function_signatures.copy()
+
     def __init__(self, namespaces=None, variable_types=None, strict=True, compatibility_mode=False,
                  default_collation=None, default_namespace=None, function_namespace=None,
                  xsd_version=None, schema=None, base_uri=None, document_types=None,
@@ -213,7 +216,9 @@ class XPath2Parser(XPath1Parser):
         if not variable_types:
             self.variable_types = {}
         elif all(self.is_sequence_type(v) for v in variable_types.values()):
-            self.variable_types = variable_types.copy()
+            self.variable_types = {
+                k: normalize_sequence_type(v) for k, v in variable_types.items()
+            }
         else:
             raise ElementPathValueError('invalid sequence type for in-scope variable types')
 
@@ -292,7 +297,7 @@ class XPath2Parser(XPath1Parser):
         return self.next_token
 
     @classmethod
-    def constructor(cls, symbol, bp=0, label='constructor function'):
+    def constructor(cls, symbol, bp=0, nargs=1, sequence_types=(), label='constructor function'):
         """Creates a constructor token class."""
         def nud_(self):
             try:
@@ -323,14 +328,19 @@ class XPath2Parser(XPath1Parser):
         def cast_(value):
             raise NotImplementedError
 
-        pattern = r'\b%s(?=\s*\(|\s*\(\:.*\:\)\()' % symbol
-        token_class = cls.register(symbol, pattern=pattern, label=label, lbp=bp, rbp=bp,
+        if not sequence_types:
+            assert nargs == 1
+            sequence_types = ('xs:anyAtomicType?', 'xs:%s?' % symbol)
+
+        token_class = cls.register(symbol, nargs=nargs, sequence_types=sequence_types,
+                                   label=label, bases=(XPathFunction,), lbp=bp, rbp=bp,
                                    nud=nud_, evaluate=evaluate_, cast=cast_)
 
         def bind(func):
-            assert func.__name__ == 'cast', \
-                "The function name must be 'cast', not %r." % func.__name__
-            setattr(token_class, func.__name__, func)
+            method_name = func.__name__.partition('_')[0]
+            if method_name != 'cast':
+                raise ValueError("The function name must be 'cast' or starts with 'cast_'")
+            setattr(token_class, method_name, func)
             return func
         return bind
 
@@ -351,16 +361,21 @@ class XPath2Parser(XPath1Parser):
             return self_
 
         def evaluate_(self_, context=None):
-            item = self_.get_argument(context)
-            if item is None:
+            arg = self_.get_argument(context)
+            if arg is None:
                 return []
-            else:
-                return self_.parser.schema.cast_as(self_[0].evaluate(context), atomic_type)
+
+            value = self_.string_value(arg)
+            try:
+                return self_.parser.schema.cast_as(value, atomic_type)
+            except (TypeError, ValueError) as err:
+                raise self_.error('FORG0001', err)
 
         symbol = get_prefixed_name(atomic_type, self.namespaces)
-        token_class_name = str("_%s_constructor_token" % symbol.replace(':', '_'))
+        token_class_name = "_%sConstructorFunction" % symbol.replace(':', '_')
         kwargs = {
             'symbol': symbol,
+            'nargs': 1,
             'label': 'constructor function',
             'pattern': r'\b%s(?=\s*\(|\s*\(\:.*\:\)\()' % symbol,
             'lbp': bp,
@@ -371,7 +386,7 @@ class XPath2Parser(XPath1Parser):
             '__qualname__': token_class_name,
             '__return__': None
         }
-        token_class = ABCMeta(token_class_name, (self.token_base_class,), kwargs)
+        token_class = ABCMeta(token_class_name, (XPathFunction,), kwargs)
         MutableSequence.register(token_class)
         self.symbol_table[symbol] = token_class
         return token_class
@@ -408,15 +423,9 @@ class XPath2Parser(XPath1Parser):
             except KeyError:
                 sequence_type = 'item()*' if isinstance(value, list) else 'item()'
 
-            if sequence_type[-1] in ('?', '+', '*'):
-                if self.match_sequence_type(value, sequence_type[:-1], sequence_type[-1]):
-                    continue
-            else:
-                if self.match_sequence_type(value, sequence_type):
-                    continue
-
-            message = "Unmatched sequence type for variable {!r}".format(varname)
-            raise xpath_error('XPDY0050', message)
+            if not self.match_sequence_type(value, sequence_type):
+                message = "Unmatched sequence type for variable {!r}".format(varname)
+                raise xpath_error('XPDY0050', message)
 
 
 ##
@@ -453,21 +462,21 @@ register(':)')
 
 @method('as')
 @method('of')
-def nud(self):
+def nud_as_and_of_symbols(self):
     raise self.error('XPDY0002')  # Dynamic context required
 
 
 ###
 # Variables
 @method('$', bp=90)
-def nud(self):
+def nud_variable_reference(self):
     self.parser.expected_name('(name)')
     self[:] = self.parser.expression(rbp=90),
     return self
 
 
 @method('$')
-def evaluate(self, context=None):
+def evaluate_variable_reference(self, context=None):
     if context is None:
         raise self.missing_context()
 
@@ -503,12 +512,12 @@ XPath2Parser.duplicate('|', 'union')
 
 @method(infix('intersect', bp=55))
 @method(infix('except', bp=55))
-def select(self, context=None):
+def select_intersect_and_except_operators(self, context=None):
     if context is None:
         raise self.missing_context()
 
     s1, s2 = set(self[0].select(copy(context))), set(self[1].select(copy(context)))
-    if any(not is_xpath_node(x) for x in s1) or any(not is_xpath_node(x) for x in s1):
+    if any(not is_xpath_node(x) for x in s1) or any(not is_xpath_node(x) for x in s2):
         raise self.error('XPTY0004', 'only XPath nodes are allowed')
 
     if self.symbol == 'except':
@@ -520,7 +529,11 @@ def select(self, context=None):
 ###
 # 'if' expression
 @method('if', bp=20)
-def nud(self):
+def nud_if_expression(self):
+    if self.parser.next_token.symbol != '(':
+        token = self.parser.symbol_table['(name)'](self.parser, self.symbol)
+        return token.nud()
+
     self.parser.advance('(')
     self[:] = self.parser.expression(5),
     self.parser.advance(')')
@@ -532,7 +545,7 @@ def nud(self):
 
 
 @method('if')
-def evaluate(self, context=None):
+def evaluate_if_expression(self, context=None):
     if self.boolean_value(self[0].evaluate(copy(context))):
         return self[1].evaluate(context)
     else:
@@ -540,7 +553,7 @@ def evaluate(self, context=None):
 
 
 @method('if')
-def select(self, context=None):
+def select_if_expression(self, context=None):
     if self.boolean_value([x for x in self[0].select(copy(context))]):
         yield from self[1].select(context)
     else:
@@ -551,8 +564,12 @@ def select(self, context=None):
 # Quantified expressions
 @method('some', bp=20)
 @method('every', bp=20)
-def nud(self):
+def nud_quantified_expressions(self):
     del self[:]
+    if self.parser.next_token.symbol != '$':
+        token = self.parser.symbol_table['(name)'](self.parser, self.symbol)
+        return token.nud()
+
     while True:
         self.parser.next_token.expected('$')
         variable = self.parser.expression(5)
@@ -564,10 +581,9 @@ def nud(self):
             if tk[0].value == variable[0].value:
                 raise tk.error('XPST0008', 'loop variable in its range expression')
 
-        if self.parser.next_token.symbol == ',':
-            self.parser.advance()
-        else:
+        if self.parser.next_token.symbol != ',':
             break
+        self.parser.advance()
 
     self.parser.advance('satisfies')
     self.append(self.parser.expression(5))
@@ -576,7 +592,7 @@ def nud(self):
 
 @method('some')
 @method('every')
-def evaluate(self, context=None):
+def evaluate_quantified_expressions(self, context=None):
     if context is None:
         raise self.missing_context()
 
@@ -599,8 +615,12 @@ def evaluate(self, context=None):
 ###
 # 'for' expressions
 @method('for', bp=20)
-def nud(self):
+def nud_for_expression(self):
     del self[:]
+    if self.parser.next_token.symbol != '$':
+        token = self.parser.symbol_table['(name)'](self.parser, self.symbol)
+        return token.nud()
+
     while True:
         self.parser.next_token.expected('$')
         variable = self.parser.expression(5)
@@ -612,10 +632,9 @@ def nud(self):
             if tk[0].value == variable[0].value:
                 raise tk.error('XPST0008', 'loop variable in its range expression')
 
-        if self.parser.next_token.symbol == ',':
-            self.parser.advance()
-        else:
+        if self.parser.next_token.symbol != ',':
             break
+        self.parser.advance()
 
     self.parser.advance('return')
     self.append(self.parser.expression(5))
@@ -623,7 +642,7 @@ def nud(self):
 
 
 @method('for')
-def select(self, context=None):
+def select_for_expression(self, context=None):
     if context is None:
         raise self.missing_context()
 
@@ -640,7 +659,7 @@ def select(self, context=None):
 # Sequence type based
 @method('instance', bp=60)
 @method('treat', bp=61)
-def led(self, left):
+def led_sequence_type_based_expressions(self, left):
     self.parser.advance('of' if self.symbol == 'instance' else 'as')
     if self.parser.next_token.label not in ('kind test', 'sequence type'):
         self.parser.expected_name('(name)', ':')
@@ -648,22 +667,18 @@ def led(self, left):
     try:
         self[:] = left, self.parser.expression(rbp=self.rbp)
     except ElementPathTypeError as err:
-        try:
-            raise err.token.wrong_syntax(err.message) from None
-        except AttributeError:
-            raise self.parser.next_token.wrong_syntax() from None
+        message = getattr(err, 'message', str(err))
+        raise self.error('XPST0003', message) from None
 
     next_symbol = self.parser.next_token.symbol
     if self[1].symbol != 'empty-sequence' and next_symbol in ('?', '*', '+'):
         self[2:] = self.parser.symbol_table[next_symbol](self.parser),  # Add nullary token
         self.parser.advance()
-    elif next_symbol in {'('}:
-        raise self.error('XPST0051')
     return self
 
 
 @method('instance')
-def evaluate(self, context=None):
+def evaluate_instance_expression(self, context=None):
     occurs = self[2].symbol if len(self) > 2 else None
     position = None
 
@@ -704,7 +719,7 @@ def evaluate(self, context=None):
 
 
 @method('treat')
-def evaluate(self, context=None):
+def evaluate_treat_expression(self, context=None):
     occurs = self[2].symbol if len(self) > 2 else None
     position = None
     castable_expr = []
@@ -713,9 +728,9 @@ def evaluate(self, context=None):
             raise self.wrong_sequence_type()
     elif self[1].label in ('kind test', 'sequence type'):
         for position, item in enumerate(self[0].select(context)):
-            if self[1].evaluate(context) is None:
-                if context is not None and not isinstance(context, XPathSchemaContext):
-                    raise self.wrong_sequence_type()
+            result = self[1].evaluate(context)
+            if isinstance(result, list) and not result:
+                raise self.wrong_sequence_type()
             elif position and (occurs is None or occurs == '?'):
                 raise self.wrong_sequence_type("more than one item in sequence")
             castable_expr.append(item)
@@ -751,8 +766,9 @@ def evaluate(self, context=None):
 # Simple type based
 @method('castable', bp=62)
 @method('cast', bp=63)
-def led(self, left):
+def led_cast_expressions(self, left):
     self.parser.advance('as')
+    self.parser.expected_name('(name)', ':')
     self[:] = left, self.parser.expression(rbp=self.rbp)
     if self.parser.next_token.symbol == '?':
         self[2:] = self.parser.symbol_table['?'](self.parser),  # Add nullary token
@@ -762,7 +778,7 @@ def led(self, left):
 
 @method('castable')
 @method('cast')
-def evaluate(self, context=None):
+def evaluate_cast_expressions(self, context=None):
     try:
         atomic_type = get_expanded_name(self[1].source, namespaces=self.parser.namespaces)
     except KeyError as err:
@@ -793,7 +809,7 @@ def evaluate(self, context=None):
     arg = self.data_value(result[0])
     try:
         if namespace != XSD_NAMESPACE:
-            value = self.parser.schema.cast_as(arg, atomic_type)
+            value = self.parser.schema.cast_as(self.string_value(arg), atomic_type)
         else:
             local_name = atomic_type.split('}')[1]
             token_class = self.parser.symbol_table.get(local_name)
@@ -803,28 +819,16 @@ def evaluate(self, context=None):
 
             token = token_class(self.parser)
             value = token.cast(arg)
+
     except ElementPathError:
         if self.symbol != 'cast':
             return False
         raise
-    except KeyError:
-        msg = "atomic type %r not found in the in-scope schema types"
-        self.unknown_atomic_type(msg % self[1].source)
-    except TypeError as err:
+    except (TypeError, ValueError) as err:
         if self.symbol != 'cast':
             return False
-        elif isinstance(arg, UntypedAtomic):
-            raise self.error('FORG0001', err)
-        elif self[0].symbol == ':' and self[0][1].symbol == 'string':
+        elif isinstance(arg, (UntypedAtomic, str)):
             raise self.error('FORG0001', err) from None
-
-        raise self.error('XPTY0004', err) from None
-    except ValueError as err:
-        if self.symbol != 'cast':
-            return False
-        elif self[0].symbol == ':' and self[0][1].symbol == 'string':
-            raise self.error('FORG0001', err) from None
-
         raise self.error('XPTY0004', err) from None
     else:
         return value if self.symbol == 'cast' else True
@@ -833,7 +837,7 @@ def evaluate(self, context=None):
 ###
 # Comma operator - concatenate items or sequences
 @method(infix(',', bp=5))
-def evaluate(self, context=None):
+def evaluate_comma_operator(self, context=None):
     results = []
     for op in self:
         result = op.evaluate(context)
@@ -845,15 +849,15 @@ def evaluate(self, context=None):
 
 
 @method(',')
-def select(self, context=None):
+def select_comma_operator(self, context=None):
     for op in self:
         yield from op.select(context=copy(context))
 
 
 ###
-# Parenthesized expressions: XPath 2.0 admits the empty case ().
-@method('(')
-def nud(self):
+# Parenthesized expression: XPath 2.0 admits the empty case ().
+@method('(', bp=100)
+def nud_parenthesized_expression(self):
     if self.parser.next_token.symbol != ')':
         self[:] = self.parser.expression(),
     self.parser.advance(')')
@@ -861,19 +865,13 @@ def nud(self):
 
 
 @method('(')
-def evaluate(self, context=None):
-    if not self:
-        return []
-    else:
-        return self[0].evaluate(context)
+def evaluate_parenthesized_expression(self, context=None):
+    return self[0].evaluate(context) if self else []
 
 
 @method('(')
-def select(self, context=None):
-    if self:
-        return self[0].select(context)
-    else:
-        return iter(())
+def select_parenthesized_expression(self, context=None):
+    return self[0].select(context) if self else iter(())
 
 
 ###
@@ -887,7 +885,7 @@ def select(self, context=None):
 @method('gt', bp=30)
 @method('le', bp=30)
 @method('ge', bp=30)
-def led(self, left):
+def led_value_comparison_operators(self, left):
     if left.symbol in {'eq', 'ne', 'lt', 'le', 'gt', 'ge'}:
         raise self.wrong_syntax()
     self[:] = left, self.parser.expression(rbp=30)
@@ -900,7 +898,7 @@ def led(self, left):
 @method('gt')
 @method('le')
 @method('ge')
-def evaluate(self, context=None):
+def evaluate_value_comparison_operators(self, context=None):
     operands = [self[0].get_atomized_operand(context=copy(context)),
                 self[1].get_atomized_operand(context=copy(context))]
 
@@ -938,24 +936,17 @@ def evaluate(self, context=None):
 ###
 # Node comparison
 @method('is', bp=30)
-def led(self, left):
+def led_node_comparison(self, left):
     if left.symbol == 'is':
         raise self.wrong_syntax()
     self[:] = left, self.parser.expression(rbp=30)
     return self
 
 
-@method('is')
-def nud(self):
-    if self.parser.next_token.symbol == '(':
-        raise self.error('XPST0017', '{} cannot have arguments'.format(self))
-    raise self.wrong_syntax()
-
-
 @method('is')
 @method(infix('<<', bp=30))
 @method(infix('>>', bp=30))
-def evaluate(self, context=None):
+def evaluate_node_comparison(self, context=None):
     symbol = self.symbol
 
     left = [x for x in self[0].select(context)]
@@ -975,7 +966,7 @@ def evaluate(self, context=None):
     else:
         if left[0] is right[0]:
             return False
-        for item in context.root.iter():
+        for item in context.root.iter():  # pragma: no cover
             if left[0] is item:
                 return True if symbol == '<<' else False
             elif right[0] is item:
@@ -987,7 +978,7 @@ def evaluate(self, context=None):
 ###
 # Range expression
 @method('to', bp=35)
-def led(self, left):
+def led_range_expression(self, left):
     if left.symbol == 'to':
         raise self.wrong_syntax()
     self[:] = left, self.parser.expression(rbp=35)
@@ -995,25 +986,23 @@ def led(self, left):
 
 
 @method('to')
-def evaluate(self, context=None):
-    start, stop = self.get_operands(context, cls=int)
+def evaluate_range_expression(self, context=None):
+    start, stop = self.get_operands(context, cls=Integer)
     try:
         return [x for x in range(start, stop + 1)]
-    except TypeError as err:
-        if context is None or start is None or stop is None:
-            return []
-        raise self.error('FORG0006', err) from None
+    except TypeError:
+        return []
 
 
 @method('to')
-def select(self, context=None):
+def select_range_expression(self, context=None):
     yield from self.evaluate(context)
 
 
 ###
 # Numerical operators
 @method(infix('idiv', bp=45))
-def evaluate(self, context=None):
+def evaluate_idiv_operator(self, context=None):
     op1, op2 = self.get_operands(context)
     if op1 is None or op2 is None:
         raise self.error('XPST0005')
@@ -1038,10 +1027,32 @@ def evaluate(self, context=None):
             return int(result) + 1
 
 
+# Resolve the intrinsic ambiguity of some infix operators
+@method('union')
+@method('intersect')
+@method('except')
+@method('eq')
+@method('ne')
+@method('lt')
+@method('gt')
+@method('le')
+@method('ge')
+@method('is')
+@method('to')
+@method('idiv')
+@method('instance')
+@method('treat')
+@method('castable')
+@method('cast')
+def nud_disambiguation_of_infix_operators(self):
+    token = self.parser.symbol_table['(name)'](self.parser, self.symbol)
+    return token.nud()
+
+
 ###
 # Kind tests (sequence types that can appear also in XPath expressions)
 @method(function('document-node', nargs=(0, 1), label='kind test'))
-def select(self, context=None):
+def select_document_node_kind_test(self, context=None):
     if context is None:
         raise self.missing_context()
     elif not self:
@@ -1057,7 +1068,7 @@ def select(self, context=None):
 
 
 @method('document-node')
-def nud(self):
+def nud_document_node_kind_test(self):
     self.parser.advance('(')
     if self.parser.next_token.symbol in ('element', 'schema-element'):
         self[0:] = self.parser.expression(5),
@@ -1071,7 +1082,7 @@ def nud(self):
 
 
 @method(function('element', nargs=(0, 2), label='kind test'))
-def select(self, context=None):
+def select_element_kind_test(self, context=None):
     if context is None:
         raise self.missing_context()
     elif not self:
@@ -1082,14 +1093,14 @@ def select(self, context=None):
         for item in self[0].select(context):
             if len(self) == 1:
                 yield item
-            elif self.xsd_types:
-                type_annotation = self[1].evaluate(context)
-                if self.xsd_types.is_matching(type_annotation, self.parser.default_namespace):
-                    yield context.item
+            elif isinstance(item, TypedElement):
+                for type_annotation in self[1].select():
+                    if type_annotation == item.xsd_type.name:
+                        yield item
 
 
 @method('element')
-def nud(self):
+def nud_element_kind_test(self):
     self.parser.advance('(')
     if self.parser.next_token.symbol != ')':
         self.parser.expected_name('(name)', ':', '*', message='a QName or a wildcard expected')
@@ -1104,7 +1115,7 @@ def nud(self):
 
 
 @method(function('schema-attribute', nargs=1, label='kind test'))
-def select(self, context=None):
+def select_schema_attribute_kind_test(self, context=None):
     if context is None:
         raise self.missing_context()
 
@@ -1123,7 +1134,7 @@ def select(self, context=None):
 
 
 @method(function('schema-element', nargs=1, label='kind test'))
-def select(self, context=None):
+def select_schema_element_kind_test(self, context=None):
     if context is None:
         raise self.missing_context()
 
@@ -1144,7 +1155,7 @@ def select(self, context=None):
 
 @method('schema-attribute')
 @method('schema-element')
-def nud(self):
+def nud_schema_node_kind_test(self):
     self.parser.advance('(')
     self.parser.expected_name('(name)', ':', message='a QName expected')
     self[0:] = self.parser.expression(5),
@@ -1167,7 +1178,7 @@ register('attribute', lbp=90, rbp=90, label=('kind test', 'axis'),
 
 
 @method('attribute')
-def nud(self):
+def nud_attribute_kind_test_or_axis(self):
     if self.parser.next_token.symbol == '::':
         self.parser.advance('::')
         self.parser.expected_name(
@@ -1189,7 +1200,7 @@ def nud(self):
 
 
 @method('attribute')
-def select(self, context=None):
+def select_attribute_kind_test_or_axis(self, context=None):
     if context is None:
         raise self.missing_context()
     elif self.label == 'axis':
@@ -1211,9 +1222,6 @@ def select(self, context=None):
                     self.add_xsd_type(attribute)
                 elif not type_name:
                     yield attribute.value
-                elif isinstance(attribute, TypedAttribute):
-                    if attribute.xsd_type.name == type_name:
-                        yield attribute.value
                 else:
                     xsd_type = self.get_xsd_type(attribute)
                     if xsd_type is not None and xsd_type.name == type_name:
