@@ -30,8 +30,10 @@ import urllib.parse
 from xml.etree.ElementTree import Element
 
 from .exceptions import ElementPathError, ElementPathValueError, XPATH_ERROR_CODES
-from .namespaces import XQT_ERRORS_NAMESPACE, XSD_NAMESPACE, XPATH_FUNCTIONS_NAMESPACE, \
-    XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE
+from .helpers import ordinal
+from .namespaces import XQT_ERRORS_NAMESPACE, XSD_NAMESPACE, \
+    XPATH_FUNCTIONS_NAMESPACE, XPATH_MATH_FUNCTIONS_NAMESPACE, \
+    XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE, XSI_NIL
 from .xpath_nodes import XPathNode, TypedElement, AttributeNode, TextNode, \
     NamespaceNode, TypedAttribute, is_etree_element, etree_iter_strings, \
     is_comment_node, is_processing_instruction_node, is_element_node, \
@@ -48,21 +50,6 @@ UNICODE_CODEPOINT_COLLATION = "http://www.w3.org/2005/xpath-functions/collation/
 XSD_SPECIAL_TYPES = {XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE}
 
 
-def ordinal(n):
-    if n in {11, 12, 13}:
-        return '%dth' % n
-
-    least_significant_digit = n % 10
-    if least_significant_digit == 1:
-        return '%dst' % n
-    elif least_significant_digit == 2:
-        return '%dnd' % n
-    elif least_significant_digit == 3:
-        return '%drd' % n
-    else:
-        return '%dth' % n
-
-
 class XPathToken(Token):
     """Base class for XPath tokens."""
     xsd_types = None  # for XPath 2.0+ schema types labeling
@@ -120,6 +107,8 @@ class XPathToken(Token):
             return '$%s' % self[0].source
         elif symbol == '{':
             return '{%s}%s' % (self[0].value, self[1].value)
+        elif symbol == 'if':
+            return 'if (%s) then %s else %s' % (self[0].source, self[1].source, self[2].source)
         elif symbol == 'instance':
             return '%s instance of %s' % (self[0].source, ''.join(t.source for t in self[1:]))
         elif symbol == 'treat':
@@ -197,7 +186,7 @@ class XPathToken(Token):
                 msg = "missing %s argument" % ordinal(index + 1)
                 raise self.error('XPST0017', msg) from None
             else:
-                return
+                return default
         else:
             item = None
             for k, result in enumerate(selector(copy(context))):
@@ -223,7 +212,7 @@ class XPathToken(Token):
                     raise self.error('XPTY0004', msg.format(ord_arg))
 
         # Type promotion checking (see "function conversion rules" in XPath 2.0 language definition)
-        if cls is not None and not isinstance(item, cls):
+        if cls is not None and not isinstance(item, cls) and not issubclass(cls, XPathToken):
             if promote and isinstance(item, promote):
                 return cls(item)
 
@@ -239,12 +228,11 @@ class XPathToken(Token):
                 value = self.data_value(item)
                 if isinstance(value, cls):
                     return value
+                elif isinstance(value, AnyURI) and issubclass(cls, str):
+                    return cls(value)
                 elif isinstance(value, UntypedAtomic):
                     try:
-                        if issubclass(cls, str):
-                            return str(value)
-                        else:
-                            return cls(value)
+                        return cls(value)
                     except (TypeError, ValueError):
                         pass
 
@@ -375,7 +363,8 @@ class XPathToken(Token):
 
         :param context: the XPath dynamic context.
         """
-        self.parser.check_variables(context.variables)
+        if context is not None:
+            self.parser.check_variables(context.variables)
 
         for result in self.select(context):
             if not isinstance(result, XPathNode):
@@ -390,7 +379,10 @@ class XPathToken(Token):
                 else:
                     yield result.value
             elif isinstance(result, NamespaceNode):  # pragma: no cover
-                yield result.prefix, result.uri
+                if self.parser.compatibility_mode:
+                    yield result.prefix, result.uri
+                else:
+                    yield result.uri
 
     def get_results(self, context):
         """
@@ -465,13 +457,15 @@ class XPathToken(Token):
 
         return op1, op2
 
-    def get_absolute_uri(self, uri, base_uri=None):
+    def get_absolute_uri(self, uri, base_uri=None, as_string=True):
         """
         Obtains an absolute URI from the argument and the static context.
 
         :param uri: a string representing an URI.
         :param base_uri: an alternative base URI, otherwise the base_uri \
         of the static context is used.
+        :param as_string: if `True` then returns the URI as a string, otherwise \
+        returns the URI as xs:anyURI instance.
         :returns: the argument if it's an absolute URI. Otherwise returns the URI
         obtained by the join o the base_uri of the static context with the
         argument. Returns the argument if the base_uri is `None'.
@@ -483,14 +477,16 @@ class XPathToken(Token):
         if url_parts.scheme or url_parts.netloc \
                 or url_parts.path.startswith('/') \
                 or base_uri is None:
-            return uri
+            return uri if as_string else AnyURI(uri)
 
         url_parts = urllib.parse.urlsplit(base_uri)
         if url_parts.fragment or not url_parts.scheme and \
                 not url_parts.netloc and not url_parts.path.startswith('/'):
             raise self.error('FORG0002', '{!r} is not suitable as base URI'.format(base_uri))
 
-        return urllib.parse.urljoin(base_uri, uri)
+        if as_string:
+            return urllib.parse.urljoin(base_uri, uri)
+        return AnyURI(urllib.parse.urljoin(base_uri, uri))
 
     def get_namespace(self, prefix):
         """
@@ -511,7 +507,7 @@ class XPathToken(Token):
         """
         if self.symbol in ('(name)', '*'):
             pass
-        elif namespace == XPATH_FUNCTIONS_NAMESPACE:
+        elif namespace == self.parser.function_namespace:
             if self.label != 'function':
                 msg = "a name, a wildcard or a function expected"
                 raise self.wrong_syntax(msg, code='XPST0017')
@@ -523,6 +519,12 @@ class XPathToken(Token):
                 raise self.wrong_syntax(msg, code='XPST0017')
             elif isinstance(self.label, MultiLabel):
                 self.label = 'constructor function'
+        elif namespace == XPATH_MATH_FUNCTIONS_NAMESPACE:
+            if self.label != 'math function':
+                msg = "a name, a wildcard or a math function expected"
+                raise self.wrong_syntax(msg, code='XPST0017')
+            elif isinstance(self.label, MultiLabel):
+                self.label = 'math function'
         else:
             raise self.wrong_syntax("a name, a wildcard or a function expected")
 
@@ -612,7 +614,7 @@ class XPathToken(Token):
                     if xsd_node.value.is_matching(name):
                         if xsd_node.name is None:
                             # node is an XSD attribute wildcard
-                            xsd_node = self.parser.schema.get_attribute(name)
+                            xsd_node = schema_context.root.maps.attributes.get(name)
                             if xsd_node is None:
                                 continue
 
@@ -623,7 +625,7 @@ class XPathToken(Token):
                 elif xsd_node.is_matching(name, self.parser.default_namespace):
                     if xsd_node.name is None:
                         # node is an XSD element wildcard
-                        xsd_node = self.parser.schema.get_element(name)
+                        xsd_node = schema_context.root.maps.elements.get(name)
                         if xsd_node is None:
                             continue
 
@@ -710,7 +712,7 @@ class XPathToken(Token):
           https://www.w3.org/TR/xpath20/#id-static-analysis
           https://www.w3.org/TR/xquery-semantics/
 
-        :param item: an untyped attribute ot element.
+        :param item: an untyped attribute or element.
         :return: a typed AttributeNode/ElementNode if the argument is matching \
         any associated XSD type.
         """
@@ -725,6 +727,8 @@ class XPathToken(Token):
                 return TypedAttribute(item, xsd_type, UntypedAtomic(item.value))
             return TypedElement(item, xsd_type, UntypedAtomic(item.text or ''))
 
+        elif isinstance(item, AttributeNode):
+            pass
         elif xsd_type.has_mixed_content():
             value = UntypedAtomic(item.text or '')
             return TypedElement(item, xsd_type, value)
@@ -732,6 +736,8 @@ class XPathToken(Token):
             return TypedElement(item, xsd_type, None)
         elif xsd_type.is_empty():
             return TypedElement(item, xsd_type, None)
+        elif item.get(XSI_NIL) and getattr(xsd_type.parent, 'nillable', None):
+            return TypedElement(item, xsd_type, None)
 
         if self.parser.xsd_version == '1.0':
             atomic_types = xsd10_atomic_types
@@ -1053,3 +1059,174 @@ class XPathToken(Token):
 
     def unknown_namespace(self, message=None):
         return self.error('XPST0081', message)
+
+
+class XPathAxis(XPathToken):
+    pattern = r'\b[^\d\W][\w.\-\xb7\u0300-\u036F\u203F\u2040]*(?=\s*\:\:|\s*\(\:.*\:\)\s*\:\:)'
+    label = 'axis'
+    reverse_axis = False
+
+    def nud(self):
+        self.parser.advance('::')
+        self.parser.expected_name(
+            '(name)', '*', 'text', 'node', 'document-node',
+            'comment', 'processing-instruction', 'attribute',
+            'schema-attribute', 'element', 'schema-element'
+        )
+        self[:] = self.parser.expression(rbp=self.rbp),
+        return self
+
+
+class ValueToken(XPathToken):
+    """
+    A dummy token for encapsulating a value.
+    """
+    symbol = '(value)'
+
+    def evaluate(self, context=None):
+        return self.value
+
+    def select(self, context=None):
+        yield self.value
+
+
+class XPathFunction(XPathToken):
+    """
+    A token for processing XPath functions.
+    """
+    _name = None
+    pattern = r'\b[^\d\W][\w.\-\xb7\u0300-\u036F\u203F\u2040]*(?=\s*(?:\(\:.*\:\))?\s*\((?!\:))'
+
+    sequence_types = ()
+    "Sequence types of arguments and of the return value of the function."
+
+    nargs = None
+    "Number of arguments: a single value or a couple with None that means unbounded."
+
+    def __init__(self, parser, nargs=None):
+        super().__init__(parser)
+        if isinstance(nargs, int) and nargs != self.nargs:
+            if nargs < 0:
+                raise self.error('XPST0017', 'number of arguments must be non negative')
+            elif isinstance(self.nargs, int) or isinstance(self.nargs, (tuple, list)) and \
+                    (self.nargs[0] > nargs or self.nargs[1] and self.nargs[1] < nargs):
+                raise self.error('XPST0017', 'incongruent number of arguments')
+            else:
+                self.nargs = nargs
+
+    def __call__(self, context=None, argument_list=None):
+        args = []
+        if isinstance(argument_list, (list, tuple)):
+            for token in argument_list:
+                args.append(token)
+        elif isinstance(argument_list, XPathToken):
+            if argument_list.symbol == '(':
+                args.append(argument_list)
+            else:
+                for token in argument_list.iter():
+                    if token.symbol not in ('(', ','):
+                        args.append(token)
+
+        context = copy(context)
+        if self.symbol == 'function':
+            if context is None:
+                raise self.missing_context()
+
+            for variable, sequence_type, value in zip(self, self.sequence_types, args):
+                if not self.parser.match_sequence_type(value, sequence_type):
+                    msg = "invalid type for argument {!r}"
+                    raise self.error('XPTY0004', msg.format(variable[0].value))
+                context.variables[variable[0].value] = value
+        elif any(tk.symbol == '?' for tk in self):
+            for value, tk in zip(args, filter(lambda x: x.symbol == '?', self)):
+                tk.value = value
+        else:
+            self.clear()
+            for value in args:
+                if isinstance(value, XPathToken):
+                    self.append(value)
+                else:
+                    self.append(ValueToken(self.parser, value=value))
+
+        result = self.evaluate(context)
+        if not self.parser.match_sequence_type(result, self.sequence_types[-1]):
+            msg = "{!r} does not match sequence type {}"
+            raise self.error('XPTY0004', msg.format(result, self.sequence_types[-1]))
+
+        return result
+
+    @property
+    def name(self):
+        if self.symbol == 'function':
+            return
+        elif self._name is None:
+            if not self.namespace or self.namespace == XPATH_FUNCTIONS_NAMESPACE:
+                self._name = QName(XPATH_FUNCTIONS_NAMESPACE, 'fn:%s' % self.symbol)
+            elif self.namespace == XSD_NAMESPACE:
+                self._name = QName(XSD_NAMESPACE, 'xs:%s' % self.symbol)
+            elif self.namespace == XPATH_MATH_FUNCTIONS_NAMESPACE:
+                self._name = QName(XPATH_MATH_FUNCTIONS_NAMESPACE, 'math:%s' % self.symbol)
+
+        return self._name
+
+    @property
+    def arity(self):
+        return self.nargs if isinstance(self.nargs, int) else len(self)
+
+    def nud(self):
+        code = 'XPST0017' if self.label == 'function' else 'XPST0003'
+        self.value = None
+        self.parser.advance('(')
+        if self.nargs is None:
+            del self[:]
+            if self.parser.next_token.symbol in (')', '(end)'):
+                raise self.error(code, 'at least an argument is required')
+            while True:
+                self.append(self.parser.expression(5))
+                if self.parser.next_token.symbol != ',':
+                    break
+                self.parser.advance()
+            self.parser.advance(')')
+            return self
+        elif self.nargs == 0:
+            if self.parser.next_token.symbol != ')':
+                if self.parser.next_token.symbol != '(end)':
+                    raise self.error(code, '%s has no arguments' % str(self))
+                raise self.parser.next_token.wrong_syntax()
+            self.parser.advance()
+            return self
+        elif isinstance(self.nargs, (tuple, list)):
+            min_args, max_args = self.nargs
+        else:
+            min_args = max_args = self.nargs
+
+        k = 0
+        while k < min_args:
+            if self.parser.next_token.symbol in (')', '(end)'):
+                msg = 'Too few arguments: expected at least %s arguments' % min_args
+                raise self.wrong_nargs(msg if min_args > 1 else msg[:-1])
+
+            self[k:] = self.parser.expression(5),
+            k += 1
+            if k < min_args:
+                if self.parser.next_token.symbol == ')':
+                    msg = 'Too few arguments: expected at least %s arguments' % min_args
+                    raise self.error(code, msg if min_args > 1 else msg[:-1])
+                self.parser.advance(',')
+
+        while max_args is None or k < max_args:
+            if self.parser.next_token.symbol == ',':
+                self.parser.advance(',')
+                self[k:] = self.parser.expression(5),
+            elif k == 0 and self.parser.next_token.symbol != ')':
+                self[k:] = self.parser.expression(5),
+            else:
+                break  # pragma: no cover
+            k += 1
+
+        if self.parser.next_token.symbol == ',':
+            msg = 'Too many arguments: expected at most %s arguments' % max_args
+            raise self.error(code, msg if max_args > 1 else msg[:-1])
+
+        self.parser.advance(')')
+        return self
