@@ -9,6 +9,7 @@
 # @author Davide Brunato <brunato@sissa.it>
 #
 import unittest
+from copy import copy
 from unittest.mock import patch
 import xml.etree.ElementTree as ElementTree
 
@@ -18,14 +19,22 @@ except ImportError:
     lxml_etree = None
 
 from elementpath import *
-from elementpath.schema_proxy import AbstractXsdType
 
 
-DummyXsdType = type(
-    'XsdType', (AbstractXsdType,),
-    dict(name=None, local_name=None, is_matching=lambda x: False, **{
-        k: lambda x: None for k in AbstractXsdType.__dict__ if k[0] != '_'
-    }))
+class DummyXsdType:
+    name = local_name = None
+
+    def is_matching(self, name, default_namespace): pass
+    def is_empty(self): pass
+    def is_simple(self): pass
+    def has_simple_content(self): pass
+    def has_mixed_content(self): pass
+    def is_element_only(self): pass
+    def is_key(self): pass
+    def is_qname(self): pass
+    def is_notation(self): pass
+    def decode(self, obj, *args, **kwargs): pass
+    def validate(self, obj, *args, **kwargs): pass
 
 
 class XPathContextTest(unittest.TestCase):
@@ -49,8 +58,8 @@ class XPathContextTest(unittest.TestCase):
     def test_copy(self):
         root = ElementTree.XML('<A><B1><C1/></B1><B2/><B3><C1/><C2/></B3></A>')
         context = XPathContext(root)
-        self.assertIsInstance(context.copy(), XPathContext)
-        self.assertIsNot(context.copy(), context)
+        self.assertIsInstance(copy(context), XPathContext)
+        self.assertIsNot(copy(context), context)
 
         context = XPathContext(root, axis='children')
         self.assertIsNone(context.copy().axis)
@@ -113,7 +122,7 @@ class XPathContextTest(unittest.TestCase):
             self.assertEqual(id(context._parent_map), parent_map_id)
 
         self.assertIsNone(context.get_parent(AttributeNode('max', '10')))
-        self.assertNotEqual(id(context._parent_map), parent_map_id)
+        self.assertEqual(id(context._parent_map), parent_map_id)
 
         parent_map_id = id(context._parent_map)
         self.assertIsNone(context.get_parent(AttributeNode('max', '10')))
@@ -141,8 +150,10 @@ class XPathContextTest(unittest.TestCase):
         self.assertEqual(context.get_path(root[2]), '/A/B3')
         self.assertEqual(context.get_path(root[2][0]), '/A/B3/C1')
         self.assertEqual(context.get_path(root[2][1]), '/A/B3/C2')
-        context._elem = root[2][1]
-        self.assertEqual(context.get_path(AttributeNode('max', '10')), '/A/B3/C2/@max')
+
+        self.assertEqual(context.get_path(AttributeNode('max', '10')), '@max')
+        attr = AttributeNode('max', '10', parent=root[2][1])
+        self.assertEqual(context.get_path(attr), '/A/B3/C2/@max')
 
         document = ElementTree.ElementTree(root)
         context = XPathContext(root=document)
@@ -155,7 +166,7 @@ class XPathContextTest(unittest.TestCase):
             self.assertEqual(context.get_path(TypedElement(root[0], xsd_type, 10)), '/A/B1')
 
         with patch.object(DummyXsdType(), 'is_simple', return_value=True) as xsd_type:
-            attr = TypedAttribute(AttributeNode('min', '1'), xsd_type, 1)
+            attr = TypedAttribute(AttributeNode('min', '1', root[1]), xsd_type, 1)
             context = XPathContext(root)
             context._elem = root[1]
             self.assertEqual(context.get_path(attr), '/A/B2/@min')
